<?php

/**
 * A class intended for integration with the VTB payment system.
 * This handler allows you to set up and process payments via VTB in your online store.
 *
 * v 1.0.1
 *
 */
class VtbPayPayment extends payment
{
    /**
     * Constants for log levels
     */
    const LOG_LEVEL_DEBUG = 'DEBUG';
    const LOG_LEVEL_ERROR = 'ERROR';
    
    /** @const string[] CURRENCIES поддерживаемые валюты */
    const CURRENCIES = ['RUB', 'RUR'];
  
    /**
     * VtbApi
     */

    /**
     * VtbApi settings and endpoints.
     * 
     * @var string|null
     */
    private $tokenUrl;
    
    /**
     * URL for creating orders in the payment system.
     *
     * @var string|null
     */
    private $orderUrl;
    
    /**
     * Mode of the payment system (test or production).
     *
     * @var string|null
     */
    private $mode;
    
    /**
     * Merchant authorization data for the client.
     *
     * @var mixed
     */
    private $clientMerchant;
    
    /**
     * Data for the client, including the grant type for OAuth.
     *
     * @var array
     */
    private $clientData = [
        'grant_type' => 'client_credentials'
    ];
    
    /**
     * Mapping of modes to their corresponding URLs.
     *
     * @var array
     */
    static private $modeUrls = [
        'PROD' => [
            'TOKEN' => 'https://open.api.vtb.ru:443/passport/oauth2/token',
            'ORDER' => 'https://gw.api.vtb.ru/openapi/smb/efcp/e-commerce/v1/orders'
        ],
        'TEST' => [
            'TOKEN' => 'https://auth.bankingapi.ru/auth/realms/kubernetes/protocol/openid-connect/token',
            'ORDER' => 'https://hackaton.bankingapi.ru/api/smb/efcp/e-commerce/api/v1/orders'
        ]
    ];
    
    
    /** 
     * Checks whether the currency of the order is supported by the payment system
     * 
     * @return bool Returns true if the order currency is supported, and false otherwise
     * @throws coreException Throws coreException when it encounters a system-level problem
     * @throws privateException Throws privateException if there's an issue with the private data
     */
    private function isSupportedCurrency() : bool 
    {
        $currency = $this->getCurrencyCode();
        return in_array(strtoupper($currency), self::CURRENCIES);
    }
    
    
    /** 
     * Checks if the order is suitable for processing by the payment system
     * 
     * @param order $order The order to check
     * @return bool Returns true if the order is suitable, and false otherwise
     */
    public function isSuitable(order $order) : bool 
    {
        return (
            $this->isSupportedCurrency() && 
            parent::isSuitable($order) && 
            ($order->getActualPrice() >= 0)
        );
    }
    
    
    /** 
     * Retrieves the customer email from a customer object
     * 
     * @param customer $customer The customer from which to retrieve the email
     * @return string The email of the customer
     * @throws publicException Throws publicException if the email is invalid
     */
    protected function getCustomerEmail(customer $customer)
    {
        $email = $customer->getEmail();

        if (
            !is_string($email) || 
            empty($email) || 
            !umiMail::checkEmail($email)
        ) {
            throw new publicException(getLabel('error-payment-wrong-customer-email'));
        }
        return $email;
    }
    
    
    /** 
     * Retrieves the customer object based on the customer ID in the order
     * 
     * @return customer Returns the customer object
     * @throws publicException Throws publicException if the customer ID is invalid
     */
    protected function getCustomer()
    {
        $customerId = $this->order->getCustomerId();
        $customerSource = selector::get('object')->id($customerId);

        if (!$customerSource instanceof iUmiObject) {
            throw new publicException(getLabel('error-payment-wrong-customer-id'));
        }
        
        return new customer($customerSource);
    }
    

    /** 
     * Processes the order by creating a payment link and redirects to it
     * 
     * @param string|null $template The template to use for processing
     * @throws Exception|publicException|privateException Throws exceptions when encountering errors during processing
     * @return null
     */
    public function process($template = null)
    {
        $order = $this->order;
        $orderId = $order->getId();
        $cmsController = cmsController::getInstance();

        try {
            // Order cost
            $amount = $this->formatPrice($order->getActualPrice());

            // Customer email
            $customer = $this->getCustomer();
            $customerEmail = $this->getCustomerEmail($customer);

            // Return Url
            $returnUrl = 'http://' . $cmsController->getCurrentDomain()->getHost() . '/emarket/gateway/' . $orderId;
            
            // Since in this CMS the order identifier does not change when the order composition changes, 
            // therefore, to create a new identifier, we associate the identifier with the price.
            $orderIdAndAmount = $orderId . '-' . $amount;

            $response = $this->setRequestData()->getOrderLink(
                $orderIdAndAmount,
                $customerEmail,
                time(),
                $amount,
                $returnUrl
            );

            $this->vtbpayLogger(__FUNCTION__ . ' > getOrderLink. ', self::LOG_LEVEL_DEBUG, [
                'arguments' => [
                    'orderIdAndAmount' => $orderIdAndAmount, 
                    'customerEmail' => $customerEmail,
                    'time' => time(), 
                    'amount' => $amount, 
                    'returnUrl' => $returnUrl
                ],
                'response' => $response
            ]);

            if (
                !isset($response['object']['payUrl']) ||
                empty($payUrl = $response['object']['payUrl'])
            ) {
                throw new publicException('Failed to get payment URL.');
            }
            
            $order->setPaymentStatus('initialized', true);
            $order->order();

            $module = $cmsController->getModule('emarket');
            if ($module) {
                $module->redirect($payUrl);
            }
            
        } catch (\Exception | publicException | privateException $e) {
            // Handle exception and log error
            $this->vtbpayLogger(sprintf(
                __FUNCTION__ . ' > VtbPay Exception: %s; Order id: (%s)', 
                $e->getMessage(), 
                $orderId
            ), self::LOG_LEVEL_ERROR);
            
            throw $e;
        } 
    }
    
    
    /** 
     * Checks the status of the payment and updates the payment status accordingly
     * 
     * @throws Exception|publicException|privateException Throws exceptions when encountering errors during polling
     */
    public function poll()
    {
        $buffer = outputBuffer::current();
        $buffer->clear();
        $buffer->contentType('text/plain');
        
        $orderIdAndAmount = getRequest('orderId');
        
        try {
            if (empty($orderIdAndAmount)) {
                throw new publicException('Error: Order ID not found.');
            }
            
            $response = $this->setRequestData()->getOrderInfo($orderIdAndAmount);
            
            $this->vtbpayLogger(__FUNCTION__ . ' > getOrderInfo. ', self::LOG_LEVEL_DEBUG, [
                'arguments' => [
                    'orderIdAndAmount' => $orderIdAndAmount,
                ],
                'response' => $response
            ]);

            if (
                isset($response['object']['status']['value']) && 
                $response['object']['status']['value'] == 'PAID'
            ) {
                $this->order->setPaymentStatus('accepted');
                $host = $this->getSuccessUrl();
            } 
            else {
                $this->order->setPaymentStatus('failed');
                $host = $this->getFailUrl();
                
                // Logging unsuccessful payment
                $this->vtbpayLogger(sprintf(
                    __FUNCTION__ . ' > getOrderInfo. Payment not paid. Order ID and amount: %s', 
                    $orderIdAndAmount
                ), self::LOG_LEVEL_ERROR);
            }
            
            $cmsController = cmsController::getInstance();
            $module = $cmsController->getModule('emarket');
            if ($module) {
                $this->order->commit();
                $module->redirect($host);
            }
        } catch (\Exception | publicException | privateException $e) {
            // Handle exception and log error
            $this->vtbpayLogger(sprintf(
                __FUNCTION__ . ' > VtbPay Exception: %s; Order ID and amount: %s', 
                $e->getMessage(), 
                $orderIdAndAmount ?: ''
            ), self::LOG_LEVEL_ERROR);
            
            throw $e;
        }
    }
    
    
    /**
     * VtbApi
     */
     
    /**
     * Sets up the necessary request data for the VTB payment API
     *
     * @return $this Returns the current instance for method chaining
     */
    private function setRequestData() 
    {
        $object = $this->object;
        
        if (!strlen($object->vtbpay_client_id) || !strlen($object->vtbpay_client_secret)) {
            throw new publicException(getLabel('error-payment-wrong-settings'));
        }
      
        $this->clientData['client_id'] = $object->vtbpay_client_id;
        $this->clientData['client_secret'] = $object->vtbpay_client_secret;
        $this->clientMerchant = (!empty($object->vtbpay_merchant_authorization)) ? $object->vtbpay_merchant_authorization : '';
        $this->mode = ($object->vtbpay_test_mode == 1) ? 'TEST': 'PROD';

        $this->tokenUrl = self::$modeUrls[$this->mode]['TOKEN'];
        $this->orderUrl = self::$modeUrls[$this->mode]['ORDER'];
        
        return $this;
    }
    

    /**
     * Retrieves the access token for the VTB payment API
     *
     * @return mixed The response from the API, which should include the access token
     */
    private function getAccessToken() 
    {
        return $this->sendRequest(
            $this->tokenUrl,
            http_build_query($this->clientData),
            [
                'Content-Type: application/x-www-form-urlencoded',
                'Merchant-Authorization: ' . $this->clientMerchant
            ]
        );
    }
    

    /**
     * Creates an order link for the specified parameters
     *
     * @param string $orderId The ID of the order
     * @param string $email The email of the customer
     * @param int $orderCreate The creation timestamp of the order
     * @param float $amount The amount of the order
     * @param string $returnUrl The URL to redirect to after the order
     * @return mixed The result from the sendRequest method, or false if the headers or merchant data is missing
     */
    private function getOrderLink(
        string $orderId,
        string $email,
        int $orderCreate,
        float $amount,
        string $returnUrl
    ) {
        $result = false;
        $args = [];
        $headers = $this->getAutorizationHeaders();
        $merchant = $this->clientMerchant;

        if ($headers) {
            $headers[] = 'Content-Type: application/json';
            if ($merchant) {
                $headers[] = 'Merchant-Authorization: ' . $merchant;
            }

            $timestamp = $orderCreate + 86400; // +1d
            $expire = date('Y-m-d\TH:i:s.v\Z', $timestamp); // ISO 8601

            $args = [
                'orderId' => $orderId,
                'orderName' => 'Заказ #' . $orderId,
                'customer' => [
                    'email' => $email
                ],
                'amount' => [
                    'value' => $amount,
                    'code' => 'RUB'
                ],
                'expire' => $expire,
                'returnUrl' => $returnUrl
            ];

            $result = $this->sendRequest(
                $this->orderUrl,
                json_encode($args),
                $headers
            );
        }

        return $result;
    }
    

    /**
     * Retrieves the information for a specific order
     *
     * @param string $orderId The ID of the order
     * @return mixed The result from the sendRequest method, or false if the headers or merchant data is missing
     */
    private function getOrderInfo(string $orderId)
    {
        $result = false;
        $headers = $this->getAutorizationHeaders();
        $merchant = $this->clientMerchant;

        if ($headers) {
            $headers[] = 'Content-Type: application/json';
            if ($merchant) {
                $headers[] = 'Merchant-Authorization: ' . $merchant;
            }

            $result = $this->sendRequest(
                $this->orderUrl . '/' . $orderId,
                json_encode([]),
                $headers,
                'GET'
            );
        }

        return $result;
    }
    

    /**
     * Generates the authorization headers for the API request
     *
     * @return array|bool The authorization headers if the access token exists, otherwise false
     */
    private function getAutorizationHeaders()
    {
       $token = $this->getAccessToken();
       
       if(!isset($token['access_token'])) {
         
           $error_description = '';
           if (
               isset($token['error_description']) && 
               !empty($token['error_description'])
           ) $error_description .= ' Error description: ' . $token['error_description'];
           
           throw new \privateException('Error getting access token.' . $error_description);
       }

       $result = [
           'X-IBM-Client-Id: ' . explode('@', $this->clientData['client_id'])[0],
           'Authorization: ' . $token['token_type'] . ' ' . $token['access_token'],
       ];

       return $result;
    }
    
    
    /**
     * Sends a HTTP request to the provided URL
     *
     * @param string $url The URL to send the request to
     * @param array $data The data to send with the request
     * @param array $headers The headers to send with the request
     * @param string $method The HTTP method to use for the request (default 'POST')
     * @param bool $json_encode Whether or not to JSON encode the data (default false)
     * @return mixed The result of the request, decoded from JSON, or false if there was an error with the request
     */
    private function sendRequest(
        $url,
        $data = [],
        $headers = [],
        $method = 'POST',
        $json_encode = false
    ) {
        $curl = curl_init();
        
        $contentType = '';
        foreach ($headers as $header) {
            if (strpos($header, 'Content-Type') !== false) {
                $contentType = $header;
                break;
            }
        }
        
        if ($method === 'POST' || $method === 'PUT' ) {
            if ($method === 'POST') {
                curl_setopt($curl, CURLOPT_POST, 1);
            } 
            else {
                curl_setopt($curl, CURLOPT_CUSTOMREQUEST, 'PUT');
            }
            
            if ($contentType === 'Content-Type: application/json' && $json_encode) {
                curl_setopt($curl, CURLOPT_POSTFIELDS, json_encode($data));
            } 
            else {
                curl_setopt($curl, CURLOPT_POSTFIELDS, $data);
            }
        } 
        else {
            $url = sprintf('%s?%s', $url, is_array($data) ? http_build_query($data): '');
        }

        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_HTTPHEADER, $headers);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false);

        $result = curl_exec($curl);
        
        if ($result === false) {
            throw new privateException('Curl error: ' . curl_error($curl));
        }
        
        if (empty($result)) {
            throw new privateException('The curl request returned an empty response');
        }
        
        $result = json_decode($result, true);
        
        if (json_last_error() != JSON_ERROR_NONE) {
            throw new privateException('Json decode error: ' . json_last_error_msg());
        }
        
        curl_close($curl);

        return $result;
    }
    
    
    /** 
     * Logs a message with optional context. Log type can be either DEBUG or ERROR.
     * 
     * @param string $message The message to log
     * @param string $type The type of log. Can be either DEBUG or ERROR
     * @param array $context An optional array of context data to log with the message
     * @param bool $print_r If set to true, the context array will be printed using print_r instead of json_encode
     * @return void
     */
    private function vtbpayLogger(
        $message, 
        $type = '',
        $context = [],
        $print_r = false
    ) {
				if (empty($type)) $type = self::LOG_LEVEL_DEBUG;
        
        if (
            empty($message) || 
            $type !== self::LOG_LEVEL_ERROR &&
            (
		            !isset($this->object->vtbpay_logging) || 
		            $this->object->vtbpay_logging !== 1
            )
        ) return null;
        
        $message = '[' . date('Y-m-d H:i:s') . '] ' . $type . ' ' . $message;
        
        if (!empty($context) && is_array($context)) {
            $message .= ' Context: '. PHP_EOL;
            if ($print_r) {
                $message .= print_r($context, true) . PHP_EOL;
            }
            else {
                $message .= json_encode($context);
            }
        }
        
        $sysLogPath = mainConfiguration::getInstance()->includeParam('sys-log-path');
        $logPath = sprintf('%s/payment/%s/', rtrim($sysLogPath, '/'), str_replace('Payment', '', get_class($this)));
        
        if (!is_dir($logPath)) {
          mkdir($logPath, 0777, true);
        }
        
        $logger = new \umiLogger($logPath);
        $logger->setFileName('vtbpay-request');
        $logger->separateByIp(false);
        $logger->push($message . PHP_EOL, false);
        $logger->save();
    }
}